"""DB (asyncdb) Extension.
DB connection for any Application.
"""
from abc import ABCMeta
from asyncdb import AsyncDB



class DBInterface(metaclass=ABCMeta):
    """
    Interface for using database connections in an Application using AsyncDB.
    """

    def get_database(
        self,
        driver: str,
        dsn: str = None,
        params: dict = None,
        timeout: int = 60,
        **kwargs
    ) -> AsyncDB:
        """Get the driver."""
        return AsyncDB(
            driver,
            dsn=dsn,
            params=params,
            timeout=timeout,
            **kwargs
        )



"""
Abstract Bot interface.
"""
from abc import ABC
import importlib
from typing import Any, List, Union, Optional
from collections.abc import Callable
import os
import uuid
import asyncio
from aiohttp import web
from langgraph.checkpoint.memory import MemorySaver
from langchain_core.vectorstores import VectorStoreRetriever
from langchain.memory import (
    ConversationBufferMemory,
    ConversationBufferWindowMemory
)
from langchain.prompts import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate
)
from langchain.retrievers import (
    EnsembleRetriever,
)
from langchain.docstore.document import Document
from langchain.chains.retrieval_qa.base import RetrievalQA
from langchain.chains.conversational_retrieval.base import (
    ConversationalRetrievalChain
)
from langchain_community.chat_message_histories import (
    RedisChatMessageHistory
)
from langchain_community.retrievers import BM25Retriever
# for exponential backoff
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)  # for exponential backoff
from datamodel.exceptions import ValidationError  # pylint: disable=E0611
from navconfig.logging import logging
from navigator_auth.conf import AUTH_SESSION_OBJECT
from ..interfaces import DBInterface
from ..conf import (
    REDIS_HISTORY_URL,
    EMBEDDING_DEFAULT_MODEL
)
## LLM configuration
from ..llms import AbstractLLM
# Vertex
try:
    from ..llms.vertex import VertexLLM
    VERTEX_ENABLED = True
except (ModuleNotFoundError, ImportError):
    VERTEX_ENABLED = False

# Google
try:
    from ..llms.google import GoogleGenAI
    GOOGLE_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GOOGLE_ENABLED = False

# Anthropic:
try:
    from ..llms.anthropic import Anthropic
    ANTHROPIC_ENABLED = True
except (ModuleNotFoundError, ImportError):
    ANTHROPIC_ENABLED = False

# OpenAI
try:
    from ..llms.openai import OpenAILLM
    OPENAI_ENABLED = True
except (ModuleNotFoundError, ImportError):
    OPENAI_ENABLED = False

# Groq
try:
    from ..llms.groq import GroqLLM
    GROQ_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GROQ_ENABLED = False
# Function for get LLM configuration.
from ..llms import get_llm

from ..utils import SafeDict
# Chat Response:
from ..models import ChatResponse
from .prompts import (
    BASIC_SYSTEM_PROMPT,
    BASIC_HUMAN_PROMPT,
    DEFAULT_BACKHISTORY
)
from .interfaces import EmptyRetriever
## Vector Stores:
from ..stores import AbstractStore, supported_stores, EmptyStore
from .retrievals import MultiVectorStoreRetriever


os.environ["GRPC_ENABLE_FORK_SUPPORT"] = "0"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # Hide TensorFlow logs if present
logging.getLogger(name='primp').setLevel(logging.INFO)
logging.getLogger(name='rquest').setLevel(logging.INFO)
logging.getLogger("grpc").setLevel(logging.CRITICAL)
logging.getLogger("tensorflow").setLevel(logging.CRITICAL)
logging.getLogger("transformers").setLevel(logging.CRITICAL)
logging.getLogger("pymilvus").setLevel(logging.INFO)

class AbstractBot(DBInterface, ABC):
    """AbstractBot.

    This class is an abstract representation a base abstraction for all Chatbots.
    """
    # TODO: make tensor and embeddings optional.
    # Define system prompt template
    system_prompt_template = BASIC_SYSTEM_PROMPT

    # Define human prompt template
    human_prompt_template = BASIC_HUMAN_PROMPT

    def __init__(
        self,
        name: str = 'Nav',
        system_prompt: str = None,
        human_prompt: str = None,
        **kwargs
    ):
        """Initialize the Chatbot with the given configuration."""
        self._request: Optional[web.Request] = None
        if system_prompt:
            self.system_prompt_template = system_prompt or BASIC_SYSTEM_PROMPT
        if human_prompt:
            self.human_prompt_template = human_prompt or BASIC_HUMAN_PROMPT
        # Chatbot ID:
        self.chatbot_id: uuid.UUID = kwargs.get(
            'chatbot_id',
            str(uuid.uuid4().hex)
        )
        # Basic Information:
        self.name: str = name
        ##  Logging:
        self.logger = logging.getLogger(
            f'{self.name}.Bot'
        )
        # Start initialization:
        self.kb = None
        self.knowledge_base: list = []
        self.return_sources: bool = kwargs.pop('return_sources', False)
        self.description = self._get_default_attr(
            'description', 'Navigator Chatbot', **kwargs
        )
        self.role = self._get_default_attr(
            'role', self.default_role(), **kwargs
        )
        self.goal = self._get_default_attr(
            'goal',
            'provide helpful information to users',
            **kwargs
        )
        self.backstory = self._get_default_attr(
            'backstory',
            default=self.default_backstory(),
            **kwargs
        )
        self.rationale = self._get_default_attr(
            'rationale',
            default=self.default_rationale(),
            **kwargs
        )
        # Definition of LLM
        self._llm_class: str = None
        self._default_llm: str = kwargs.get('use_llm', 'vertexai')
        self._llm_model = kwargs.get('model_name', 'gemini-2.0-pro')
        self._llm_config = kwargs.get('model_config', {})
        if self._llm_config:
            self._llm_model = self._llm_config.pop('model', self._llm_model)
            self._llm_class = self._llm_config.pop('name', None)
        # Overrriding LLM object
        self._llm_obj: Callable = kwargs.get('llm', None)
        # LLM base Object:
        self._llm: Callable = None
        self.context = kwargs.pop('context', '')

        # Pre-Instructions:
        self.pre_instructions: list = kwargs.get(
            'pre_instructions',
            []
        )

        # Knowledge base:
        self.knowledge_base: list = []
        self._documents_: list = []
        # Models, Embed and collections
        # Vector information:
        self._use_vector: bool = kwargs.get('use_vectorstore', False)
        self._vector_info_: dict = kwargs.get('vector_info', {})
        self._vector_store: dict = kwargs.get('vector_store', None)
        self.chunk_size: int = int(kwargs.get('chunk_size', 2048))
        self.dimension: int = int(kwargs.get('dimension', 768))
        self.store: Callable = None
        self.stores: List[AbstractStore] = []
        self.memory: Callable = None
        # Embedding Model Name
        self.embedding_model = kwargs.get(
            'embedding_model',
            {
                'model_name': EMBEDDING_DEFAULT_MODEL,
                'model_type': 'huggingface'
            }
        )
        # embedding object:
        self.embeddings = kwargs.get('embeddings', None)
        self.rag_model = kwargs.get(
            'rag_model',
            "rlm/rag-prompt-llama"
        )
        # Summarization and Classification Models
        # Bot Security and Permissions:
        _default = self.default_permissions()
        _permissions = kwargs.get('permissions', _default)
        if _permissions is None:
            _permissions = {}
        self._permissions = {**_default, **_permissions}

    def default_permissions(self) -> dict:
        """
        Returns the default permissions for the bot.

        This function defines and returns a dictionary containing the default
        permission settings for the bot. These permissions are used to control
        access and functionality of the bot across different organizational
        structures and user groups.

        Returns:
            dict: A dictionary containing the following keys, each with an empty list as its value:
                - "organizations": List of organizations the bot has access to.
                - "programs": List of programs the bot is allowed to interact with.
                - "job_codes": List of job codes the bot is authorized for.
                - "users": List of specific users granted access to the bot.
                - "groups": List of user groups with bot access permissions.
        """
        return {
            "organizations": [],
            "programs": [],
            "job_codes": [],
            "users": [],
            "groups": [],
        }

    def permissions(self):
        return self._permissions

    def _get_default_attr(self, key, default: Any = None, **kwargs):
        if key in kwargs:
            return kwargs.get(key)
        if hasattr(self, key):
            return getattr(self, key)
        if not hasattr(self, key):
            return default
        return getattr(self, key)

    def __repr__(self):
        return f"<Bot.{self.__class__.__name__}:{self.name}>"

    def default_rationale(self) -> str:
        # TODO: read rationale from a file
        return (
            "When responding to user queries, ensure that you provide accurate and up-to-date information.\n"
            "Be polite and clear in your explanations, "
            "ensuring that responses are based only on verified information from owned sources. "
            "If you are unsure, let the user know and avoid making assumptions. Maintain a professional tone in all responses.\n"
            "You are a fluent speaker, you can talk and respond fluently in English or Spanish, and you must answer in the same language as the user's question. If the user's language is not English, you should translate your response into their language.\n"
        )

    def default_backstory(self) -> str:
        return (
            DEFAULT_BACKHISTORY
        )

    def default_role(self) -> str:
        return "Assisting with queries"

    @property
    def llm(self):
        return self._llm

    @llm.setter
    def llm(self, model):
        self._llm_obj = model
        self._llm = model.get_llm()

    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    def llm_chain(
        self,
        llm: str = "vertexai",
        **kwargs
    ) -> AbstractLLM:
        """llm_chain.

        Args:
            llm (str): The language model to use.

        Returns:
            AbstractLLM: The language model to use.

        """
        if llm == 'openai' and OPENAI_ENABLED:
            mdl = OpenAILLM(model="gpt-3.5-turbo", **kwargs)
        elif llm in ('vertexai', 'VertexLLM') and VERTEX_ENABLED:
            mdl = VertexLLM(model="gemini-1.5-pro", **kwargs)
        elif llm == 'anthropic' and ANTHROPIC_ENABLED:
            mdl = Anthropic(model="claude-3-opus-20240229", **kwargs)
        elif llm in ('groq', 'Groq') and GROQ_ENABLED:
            mdl = GroqLLM(model="llama3-70b-8192", **kwargs)
        elif llm == 'llama3' and GROQ_ENABLED:
            mdl = GroqLLM(model="llama3-groq-70b-8192-tool-use-preview", **kwargs)
        elif llm == 'mixtral' and GROQ_ENABLED:
            mdl = GroqLLM(model="mixtral-8x7b-32768", **kwargs)
        elif llm == 'google' and GOOGLE_ENABLED:
            mdl = GoogleGenAI(model="models/gemini-1.5-pro-latest", **kwargs)
        else:
            raise ValueError(f"Invalid llm: {llm}")
        # get the LLM:
        return mdl

    def configure_llm(
        self,
        llm: Union[str, Callable] = None,
        config: Optional[dict] = None,
        use_chat: bool = False
    ):
        """
        Configuration of LLM.
        """
        if not llm:
            # Using Configuration or Default
            if self._llm_class:
                self._llm_obj = get_llm(
                    llm_name=self._llm_class,
                    model_name=self._llm_model,
                    **self._llm_config
                )
            else:
                # Using Default Configuration:
                self._llm_obj = self.llm_chain(
                    llm=self._default_llm,
                    temperature=0.2,
                    top_k=30,
                    Top_p=0.6,
                )
            self._llm = self._llm_obj.get_llm()
        elif isinstance(llm, str):
            # Get the LLM By Name:
            self._llm_obj = self.llm_chain(
                llm,
                **config
            )
            # getting langchain LLM from Obj:
            self._llm = self._llm_obj.get_llm()
        elif isinstance(llm, AbstractLLM):
            self._llm_obj = llm
            self._llm = llm.get_llm()
        elif isinstance(self._llm_obj, str):
            # is the name of the LLM object to be used:
            self._llm_obj = self.llm_chain(
                llm=self._llm_obj
            )
            self._llm = self._llm_obj.get_llm()
        elif isinstance(self._llm_obj, AbstractLLM):
            self._llm = self._llm_obj.get_llm()
        elif self._llm_obj is not None:
            self._llm = self._llm_obj
        else:
            # TODO: Calling a Default LLM
            # TODO: passing the default configuration
            self._llm_obj = self.llm_chain(
                llm=self._default_llm,
                temperature=0.2,
                top_k=30,
                top_p=0.6,
                use_chat=use_chat
            )
            self._llm = self._llm_obj.get_llm()

    def create_kb(self, documents: list):
        new_docs = []
        for doc in documents:
            content = doc.pop('content')
            source = doc.pop('source', 'knowledge-base')
            if doc:
                meta = {
                    'source': source,
                    **doc
                }
            else:
                meta = {'source': source}
            if content:
                new_docs.append(
                    Document(
                        page_content=content,
                        metadata=meta
                    )
                )
        return new_docs

    def _define_prompt(self, config: Optional[dict] = None):
        """
        Define the System Prompt and replace variables.
        """
        # setup the prompt variables:
        if config:
            for key, val in config.items():
                setattr(self, key, val)
        # Creating the variables:
        self.system_prompt_template = self.system_prompt_template.format_map(
            SafeDict(
                name=self.name,
                role=self.role,
                goal=self.goal,
                backstory=self.backstory,
                rationale=self.rationale
            )
        )
        # print('Template Prompt: \n', self.system_prompt_template)

    async def configure(self, app=None) -> None:
        """Basic Configuration of Bot.
        """
        self.app = None
        if app:
            if isinstance(app, web.Application):
                self.app = app  # register the app into the Extension
            else:
                self.app = app.get_app()  # Nav Application
        # adding this configured chatbot to app:
        if self.app:
            self.app[f"{self.name.lower()}_bot"] = self
        # Configure LLM:
        self.configure_llm()
        # And define Prompt:
        self._define_prompt()
        # Configure VectorStore if enabled:
        if self._use_vector:
            self.configure_store()

    def _get_database_store(self, store: dict) -> AbstractStore:
        name = store.get('name', 'milvus')
        store_cls = supported_stores.get(name)
        cls_path = f"parrot.stores.{name}"
        try:
            module = importlib.import_module(cls_path, package=name)
            store_cls = getattr(module, store_cls)
            return store_cls(
                embedding_model=self.embedding_model,
                embedding=self.embeddings,
                **store
            )
        except (ModuleNotFoundError, ImportError) as e:
            self.logger.error(
                f"Error importing VectorStore: {e}"
            )
            raise

    def configure_store(self, **kwargs):
        # TODO: Implement VectorStore Configuration
        print('SELF > ', self._vector_store)
        if isinstance(self._vector_store, list):
            # Is a list of vector stores instances:
            for st in self._vector_store:
                try:
                    store_cls = self._get_database_store(st)
                    store_cls.use_database = self._use_vector
                    self.stores.append(store_cls)
                except ImportError:
                    continue
        elif isinstance(self._vector_store, dict):
            print('< ENTERING HERE > ', )
            # Is a single vector store instance:
            store_cls = self._get_database_store(self._vector_store)
            store_cls.use_database = self._use_vector
            self.stores.append(store_cls)
        else:
            raise ValueError(
                f"Invalid Vector Store Config: {self._vector_store}"
            )
        self.logger.info(f"Configured Vector Stores: {self.stores}")
        if self.stores:
            self.store = self.stores[0]
        print('=================================')
        print('END STORES >> ', self.stores, self.store)
        print('=================================')


    def get_memory(
        self,
        session_id: str = None,
        key: str = 'chat_history',
        input_key: str = 'question',
        output_key: str = 'answer',
        size: int = 5,
        ttl: int = 86400
    ):
        args = {
            'memory_key': key,
            'input_key': input_key,
            'output_key': output_key,
            'return_messages': True,
            'max_len': size,
            'k': 10
        }
        if session_id:
            message_history = RedisChatMessageHistory(
                url=REDIS_HISTORY_URL,
                session_id=session_id,
                ttl=ttl
            )
            args['chat_memory'] = message_history
        return ConversationBufferWindowMemory(
            **args
        )

    def clean_history(
        self,
        session_id: str = None
    ):
        try:
            redis_client = RedisChatMessageHistory(
                url=REDIS_HISTORY_URL,
                session_id=session_id,
                ttl=60
            )
            redis_client.clear()
        except Exception as e:
            self.logger.error(
                f"Error clearing chat history: {e}"
            )

    def get_response(self, response: dict):
        if 'error' in response:
            return response  # return this error directly
        try:
            response = ChatResponse(**response)
            response.response = self.as_markdown(
                response,
                return_sources=self.return_sources
            )
            return response
        except (ValueError, TypeError) as exc:
            self.logger.error(
                f"Error validating response: {exc}"
            )
            return response
        except ValidationError as exc:
            self.logger.error(
                f"Error on response: {exc.payload}"
            )
            return response

    async def conversation(
            self,
            question: str,
            chain_type: str = 'stuff',
            search_type: str = 'similarity',
            search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
            return_docs: bool = True,
            metric_type: str = None,
            memory: Any = None,
            **kwargs
    ):
        # re-configure LLM:
        new_llm = kwargs.pop('llm', None)
        llm_config = kwargs.pop(
            'llm_config',
            {
                "temperature": 0.2,
                "top_k": 30,
                "Top_p": 0.6
            }
        )
        if new_llm:
            self.configure_llm(llm=new_llm, config=llm_config)
        # define the Pre-Context
        pre_context = "\n".join(f"- {a}." for a in self.pre_instructions)
        custom_template = self.system_prompt_template.format_map(
            SafeDict(
                summaries=pre_context
            )
        )

        # Create prompt templates
        self.system_prompt = SystemMessagePromptTemplate.from_template(
            custom_template
        )
        self.human_prompt = HumanMessagePromptTemplate.from_template(
            self.human_prompt_template,
            input_variables=['question', 'chat_history']
        )
        # Combine into a ChatPromptTemplate
        chat_prompt = ChatPromptTemplate.from_messages([
            self.system_prompt,
            self.human_prompt
        ])
        if not memory:
            memory = self.memory
        if not self.memory:
            # Initialize memory
            self.memory = ConversationBufferMemory(
                memory_key="chat_history",
                input_key="question",
                output_key='answer',
                return_messages=True
            )
        try:
            if self._use_vector:
                async with self.store as store:  #pylint: disable=E1101
                    vector = store.get_vector(metric_type=metric_type)
                    retriever = VectorStoreRetriever(
                        vectorstore=vector,
                        search_type=search_type,
                        chain_type=chain_type,
                        search_kwargs=search_kwargs
                    )
                    # Create the ConversationalRetrievalChain with custom prompt
                    chain = ConversationalRetrievalChain.from_llm(
                        llm=self._llm,
                        retriever=retriever,
                        memory=self.memory,
                        chain_type=chain_type,  # e.g., 'stuff', 'map_reduce', etc.
                        verbose=True,
                        return_source_documents=return_docs,
                        return_generated_question=True,
                        combine_docs_chain_kwargs={
                            'prompt': chat_prompt
                        },
                        **kwargs
                    )
                    response = await chain.ainvoke(
                        {"question": question}
                    )
            else:
                retriever = EmptyRetriever()
                # Create the ConversationalRetrievalChain with custom prompt
                chain = ConversationalRetrievalChain.from_llm(
                    llm=self._llm,
                    retriever=retriever,
                    memory=self.memory,
                    chain_type=chain_type,  # e.g., 'stuff', 'map_reduce', etc.
                    verbose=True,
                    return_source_documents=False,
                    return_generated_question=True,
                    combine_docs_chain_kwargs={
                        'prompt': chat_prompt
                    },
                    **kwargs
                )
                response = await chain.ainvoke(
                    {"question": question}
                )
        except asyncio.CancelledError:
            # Handle task cancellation
            print("Conversation task was cancelled.")
        except Exception as e:
            self.logger.error(
                f"Error in conversation: {e}"
            )
            raise
        return self.get_response(response)

    async def question(
            self,
            question: str,
            chain_type: str = 'stuff',
            search_type: str = 'similarity',
            search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
            return_docs: bool = True,
            metric_type: str = None,
            **kwargs
    ):
        pre_context = "\n".join(f"- {a}." for a in self.pre_instructions)
        system_prompt = self.system_prompt_template.format_map(
            SafeDict(
                summaries=pre_context
            )
        )
        human_prompt = self.human_prompt_template.replace(
            '**Chat History:**', ''
        )
        human_prompt = human_prompt.format_map(
            SafeDict(
                chat_history=''
            )
        )
        # re-configure LLM:
        new_llm = kwargs.pop('llm', None)
        llm_config = kwargs.pop(
            'llm_config',
            {
                "temperature": 0.2,
                "top_k": 30,
                "Top_p": 0.6
            }
        )
        if new_llm:
            self.configure_llm(llm=new_llm, config=llm_config)
        # Combine into a ChatPromptTemplate
        prompt = PromptTemplate(
            template=system_prompt + '\n' + human_prompt,
            input_variables=['context', 'question']
        )
        try:
            if self._use_vector:
                async with self.store as store:  #pylint: disable=E1101
                    vector = store.get_vector(metric_type=metric_type)
                    retriever = VectorStoreRetriever(
                        vectorstore=vector,
                        search_type=search_type,
                        chain_type=chain_type,
                        search_kwargs=search_kwargs
                    )
                    chain = RetrievalQA.from_chain_type(
                        llm=self._llm,
                        chain_type=chain_type,  # e.g., 'stuff', 'map_reduce', etc.
                        retriever=retriever,
                        chain_type_kwargs={
                            'prompt': prompt,
                        },
                        return_source_documents=return_docs,
                        **kwargs
                    )
                    response = await chain.ainvoke(
                        question
                    )
            else:
                retriever = EmptyRetriever()
                # Create the RetrievalQA chain with custom prompt
                chain = RetrievalQA.from_chain_type(
                    llm=self._llm,
                    chain_type=chain_type,  # e.g., 'stuff', 'map_reduce', etc.
                    retriever=retriever,
                    chain_type_kwargs={
                        'prompt': prompt,
                    },
                    return_source_documents=return_docs,
                    **kwargs
                )
                response = await chain.ainvoke(
                    question
                )
        except (RuntimeError, asyncio.CancelledError):
            # check for "Event loop is closed"
            response = chain.invoke(
                question
            )
        except Exception as e:
            # Handle exceptions
            self.logger.error(
                f"An error occurred: {e}"
            )
            response = {
                "query": question,
                "error": str(e)
            }
        return self.get_response(response)

    def as_markdown(self, response: ChatResponse, return_sources: bool = False) -> str:
        markdown_output = f"**Question**: {response.question}  \n"
        markdown_output += f"**Answer**: {response.answer}  \n"
        if return_sources is True and response.source_documents:
            source_documents = response.source_documents
            current_sources = []
            block_sources = []
            count = 0
            d = {}
            for source in source_documents:
                if count >= 20:
                    break  # Exit loop after processing 10 documents
                metadata = source.metadata
                if 'url' in metadata:
                    src = metadata.get('url')
                elif 'filename' in metadata:
                    src = metadata.get('filename')
                else:
                    src = metadata.get('source', 'unknown')
                if src == 'knowledge-base':
                    continue  # avoid attaching kb documents
                source_title = metadata.get('title', src)
                if source_title in current_sources:
                    continue
                current_sources.append(source_title)
                if src:
                    d[src] = metadata.get('document_meta', {})
                source_filename = metadata.get('filename', src)
                if src:
                    block_sources.append(f"- [{source_title}]({src})")
                else:
                    if 'page_number' in metadata:
                        block_sources.append(
                            f"- {source_filename} (Page {metadata.get('page_number')})"
                        )
                    else:
                        block_sources.append(f"- {source_filename}")
            if block_sources:
                markdown_output += f"**Sources**:  \n"
                markdown_output += "\n".join(block_sources)
            if d:
                response.documents = d
        return markdown_output

    async def __aenter__(self):
        if not self.store:
            self.store = EmptyStore()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    def retrieval(self, request: web.Request = None) -> "AbstractBot":
        """
        Configure the retrieval chain for the Chatbot, returning `self` if allowed,
        or raise HTTPUnauthorized if not. A permissions dictionary can specify
        * users
        * groups
        * job_codes
        * programs
        * organizations
        If a permission list is the literal string "*", it means "unrestricted" for that category.

        Args:
            request (web.Request, optional): The request object. Defaults to None.
        Returns:
            AbstractBot: The Chatbot object or raise HTTPUnauthorized.
        """
        self._request = request
        session = request.session
        try:
            userinfo = session[AUTH_SESSION_OBJECT]
        except KeyError:
            userinfo = {}

        # decode your user from session
        try:
            user = session.decode("user")
        except (KeyError, TypeError):
            raise web.HTTPUnauthorized(
                reason="Invalid user"
            )

        # 1: superuser is always allowed
        if userinfo.get('superuser', False) is True:
            return self

        # convenience references
        users_allowed = self._permissions.get('users', [])
        groups_allowed = self._permissions.get('groups', [])
        job_codes_allowed = self._permissions.get('job_codes', [])
        programs_allowed = self._permissions.get('programs', [])
        orgs_allowed = self._permissions.get('organizations', [])

        # 2: check if 'users' == "*" or user.username in 'users'
        if users_allowed == "*":
            return self
        if user.get('username') in users_allowed:
            return self

        # 3: check job_code
        if job_codes_allowed == "*":
            return self
        try:
            if user.job_code in job_codes_allowed:
                return self
        except AttributeError:
            pass

        # 4: check groups
        # If groups_allowed == "*", no restriction on groups
        if groups_allowed == "*":
            return self
        # otherwise, see if there's an intersection
        user_groups = set(userinfo.get("groups", []))
        if not user_groups.isdisjoint(groups_allowed):
            return self

        # 5: check programs
        if programs_allowed == "*":
            return self
        try:
            user_programs = set(userinfo.get("programs", []))
            if not user_programs.isdisjoint(programs_allowed):
                return self
        except AttributeError:
            pass


        # 6: check organizations
        if orgs_allowed == "*":
            return self
        try:
            user_orgs = set(userinfo.get("organizations", []))
            if not user_orgs.isdisjoint(orgs_allowed):
                return self
        except AttributeError:
            pass

        # If none of the conditions pass, raise unauthorized:
        raise web.HTTPUnauthorized(
            reason=f"User {user.username} is not Unauthorized"
        )

    async def invoke(
        self,
        question: str,
        chain_type: str = 'stuff',
        search_type: str = 'similarity',
        search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
        return_docs: bool = True,
        metric_type: str = None,
        memory: Any = None,
        **kwargs
    ) -> ChatResponse:
        """Build a Chain to answer Questions using AI Models.
        """
        new_llm = kwargs.pop('llm', None)
        if new_llm is not None:
            # re-configure LLM:
            llm_config = kwargs.pop(
                'llm_config',
                {
                    "temperature": 0.2,
                    "top_k": 30,
                    "Top_p": 0.6
                }
            )
            self.configure_llm(llm=new_llm, config=llm_config)
        # define the Pre-Context
        pre_context = "\n".join(f"- {a}." for a in self.pre_instructions)
        custom_template = self.system_prompt_template.format_map(
            SafeDict(
                summaries=pre_context
            )
        )
        # Create prompt templates
        self.system_prompt = SystemMessagePromptTemplate.from_template(
            custom_template
        )
        self.human_prompt = HumanMessagePromptTemplate.from_template(
            self.human_prompt_template,
            input_variables=['question', 'chat_history']
        )
        # Combine into a ChatPromptTemplate
        chat_prompt = ChatPromptTemplate.from_messages([
            self.system_prompt,
            self.human_prompt
        ])
        if not memory:
            memory = self.memory
        if not self.memory:
            # Initialize memory
            self.memory = ConversationBufferMemory(
                memory_key="chat_history",
                input_key="question",
                output_key='answer',
                return_messages=True
            )
        async with self.store as store:  #pylint: disable=E1101
            print('STORES > ', self.stores, store)
            if self._use_vector:
                if len(self.stores) > 1:
                    retriever = MultiVectorStoreRetriever(
                        stores=self.stores,
                        metric_type=metric_type,
                        search_type=search_type,
                        chain_type=chain_type,
                        search_kwargs=search_kwargs
                    )
                else:
                    vector = store.get_vector(metric_type=metric_type)
                    retriever = VectorStoreRetriever(
                        vectorstore=vector,
                        search_type=search_type,
                        chain_type=chain_type,
                        search_kwargs=search_kwargs
                    )
            else:
                retriever = EmptyRetriever()
            if self.kb:
                b25_retriever = BM25Retriever.from_documents(self.kb)
                retriever = EnsembleRetriever(
                    retrievers=[retriever, b25_retriever],
                    weights=[0.8, 0.2]
                )
            try:
                # Create the ConversationalRetrievalChain with custom prompt
                chain = ConversationalRetrievalChain.from_llm(
                    llm=self._llm,
                    retriever=retriever,
                    memory=self.memory,
                    chain_type=chain_type,  # e.g., 'stuff', 'map_reduce', etc.
                    verbose=True,
                    return_source_documents=return_docs,
                    return_generated_question=True,
                    combine_docs_chain_kwargs={
                        'prompt': chat_prompt
                    },
                    **kwargs
                )
                response = await chain.ainvoke(
                    {"question": question}
                )
                return self.get_response(response)
            except asyncio.CancelledError:
                # Handle task cancellation
                print("Conversation task was cancelled.")
            except Exception as e:
                self.logger.error(
                    f"Error in conversation: {e}"
                )
                raise



from typing import Union
from pathlib import Path, PurePath
from pydantic import PrivateAttr
from crewai_tools import BaseTool


class SaveFile(BaseTool):
    """Save a file to a directory."""
    name: str = "Save File"
    description: str = "Save a file to a directory."
    _directory: PrivateAttr

    def __init__(self, directory: Union[str, PurePath], **kwargs):
        super().__init__(**kwargs)
        if isinstance(directory, str):
            self._directory = Path(directory).resolve()
        self._directory = directory

    def _run(self, file: str, **kwargs) -> dict:
        """Run the Save File Tool."""
        filename = self._directory.joinpath(file)
        with open(filename, "w") as f:
            f.write(kwargs.get("content", ""))
        return {"file": file}



from langchain_core.prompts import ChatPromptTemplate
from langchain_groq import ChatGroq
from navconfig import config
from .abstract import AbstractLLM


class GroqLLM(AbstractLLM):
    """GroqLLM.
        Using Groq Open-source models.
    """
    model: str = "mixtral-8x7b-32768"
    supported_models: list = [
        "llama-3.3-70b-versatile",
        "qwen-2.5-32b",
        "qwen-2.5-coder-32b",
        "deepseek-r1-distill-qwen-32b",
        "deepseek-r1-distill-llama-70b",
        "gemma2-9b-it",
        "llama3-70b-8192",
        "llama3-80b-8192",
        "llama-guard-3-8b",
        "llama-3.1-8b-instant",
        "mistral-saba-24b",
        "mixtral-8x7b-32768",
        "whisper-large-v3",
        "whisper-large-v3-turbo",
    ]

    def __init__(self, *args, **kwargs):
        self.model_type = kwargs.get("model_type", "text")
        system = kwargs.pop('system_prompt', "You are a helpful assistant.")
        human = kwargs.pop('human_prompt', "{question}")
        super().__init__(*args, **kwargs)
        self._api_key = kwargs.pop('api_key', config.get('GROQ_API_KEY'))
        self._llm = ChatGroq(
            model_name=self.model,
            groq_api_key=self._api_key,
            temperature=self.temperature,
            max_retries=self.max_retries,
            max_tokens=self.max_tokens,
            model_kwargs={
                "top_p": self.top_p,
                # "top_k": self.top_k,
            },
        )
        self._embed = None # Not supported
        self.prompt = ChatPromptTemplate.from_messages(
            [("system", system), ("human", human)]
        )



from langchain_openai import (  # pylint: disable=E0401, E0611
    OpenAI,
    ChatOpenAI,
)
from navconfig import config
from .abstract import AbstractLLM


class OpenAILLM(AbstractLLM):
    """OpenAI.
    Interact with OpenAI Language Model.

    Returns:
        _type_: an instance of OpenAI LLM Model.
    """
    model: str = "gpt-4-turbo"
    max_tokens: int = 1024
    supported_models: list = [
        'gpt-4o-mini',
        'gpt-4-turbo',
        'gpt-4-turbo-preview',
        'gpt-4o',
        'gpt-3.5-turbo',
        'gpt-3.5-turbo-instruct',
        'dall-e-3'
        'tts-1',
    ]

    def __init__(self, *args, **kwargs):
        self.model_type = kwargs.get("model_type", "text")
        super().__init__(*args, **kwargs)
        self.model = kwargs.get("model", "davinci")
        self._api_key = kwargs.pop('api_key', config.get('OPENAI_API_KEY'))
        organization = config.get("OPENAI_ORGANIZATION")
        if self.model_type == 'chat':
            base_llm = ChatOpenAI
        else:
            base_llm = OpenAI
        self._llm = base_llm(
            model_name=self.model,
            api_key=self._api_key,
            organization=organization,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            **self.args
        )



import requests
from langchain.tools import BaseTool
from langchain.tools import Tool
from langchain_community.utilities import OpenWeatherMapAPIWrapper
from navconfig import config
import orjson

class OpenWeatherMapTool(BaseTool):
    """Tool that searches the OpenWeatherMap API."""
    name: str = "OpenWeatherMap"
    description: str = (
        "A wrapper around OpenWeatherMap. "
        "Useful for when you need to answer general questions about "
        "weather, temperature, humidity, wind speed, or other weather-related information. "
    )
    search: Tool = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.search = OpenWeatherMapAPIWrapper(
            openweathermap_api_key=config.get('OPENWEATHER_APPID')
        )

    def _run(
        self,
        query: dict,
    ) -> dict:
        """Use the OpenWeatherMap tool."""
        return self.search.run(query)


class OpenWeather(BaseTool):
    """
    Tool to get weather information about a location.
    """
    name: str = 'openweather_tool'
    description: str = (
        "Get weather information about a location, use this tool to answer questions about weather or weather forecast."
        " Input should be the latitude and longitude of the location you want weather information about."
    )
    base_url: str = 'http://api.openweathermap.org/'
    units: str = 'metric'
    days: int = 3
    appid: str = None
    request: str = 'weather'
    country: str = 'us'


    def __init__(self, request: str = 'weather', country: str = 'us', **kwargs):
        super().__init__(**kwargs)
        self.request = request
        self.country = country
        self.appid = config.get('OPENWEATHER_APPID')

    def _run(self, query: dict) -> dict:
        q = orjson.loads(query)  # pylint: disable=no-member
        if 'latitude' in q and 'longitude' in q:
            lat = q['latitude']
            lon = q['longitude']
            if self.request == 'weather':
                url = f"{self.base_url}data/2.5/weather?lat={lat}&lon={lon}&units={self.units}&appid={self.appid}"
            elif self.request == 'forecast':
                url = f"{self.base_url}data/2.5/forecast?lat={lat}&lon={lon}&units={self.units}&cnt={self.days}&appid={self.appid}"
        else:
            return {'error': 'Latitude and longitude are required'}
        response = requests.get(url)
        return response.json()

    async def _arun(self, query: dict) -> dict:
        raise NotImplementedError("Async method not implemented yet")



"""
Chatbot Manager.

Tool for instanciate, managing and interacting with Chatbot through APIs.
"""
from typing import Any, Dict, Type
from importlib import import_module
from aiohttp import web
from navconfig.logging import logging
from .bots.abstract import AbstractBot
from .bots.basic import BasicBot
from .bots.chatbot import Chatbot
from .handlers.chat import ChatHandler, BotHandler
from .handlers import ChatbotHandler
from .models import ChatbotModel


class BotManager:
    """BotManager.

    Manage Bots/Agents and interact with them through via aiohttp App.
    Deploy and manage chatbots and agents using a RESTful API.

    """
    app: web.Application = None

    def __init__(self) -> None:
        self.app = None
        self._bots: Dict[str, AbstractBot] = {}
        self.logger = logging.getLogger(
            name='Parrot.Manager'
        )

    def get_bot_class(self, class_name: str) -> Type[AbstractBot]:
        """
        Dynamically import a Bot class based on the class name
        from the relative module '.bots'.
        Args:
        class_name (str): The name of the Bot class to be imported.
        Returns:
        Type[AbstractBot]: A Bot class derived from AbstractBot.
        """
        module = import_module('.bots', __package__)
        try:
            return getattr(module, class_name)
        except AttributeError:
            raise ImportError(
                f"No class named '{class_name}' found in the module 'bots'."
            )

    async def load_bots(self, app: web.Application) -> None:
        """Load all chatbots from DB."""
        self.logger.info("Loading chatbots from DB...")
        db = app['database']
        async with await db.acquire() as conn:
            ChatbotModel.Meta.connection = conn
            try:
                bots = await ChatbotModel.filter(enabled=True)
            except Exception as e:
                self.logger.error(
                    f"Failed to load chatbots from DB: {e}"
                )
                return
            for bot in bots:
                if bot.bot_type == 'chatbot':
                    self.logger.notice(
                        f"Loading chatbot '{bot.name}'..."
                    )
                    cls_name = bot.bot_class
                    if cls_name is None:
                        class_name = Chatbot
                    else:
                        class_name = self.get_bot_class(cls_name)
                    chatbot = class_name(
                        chatbot_id=bot.chatbot_id,
                        name=bot.name,
                        description=bot.description,
                        use_llm=bot.llm,
                        model_name=bot.model_name,
                        model_config=bot.model_config,
                        embedding_model=bot.embedding_model,
                        use_vectorstore=bot.vector_store,
                        vector_store=bot.database,
                        config_file=bot.config_file,
                        role=bot.role,
                        goal=bot.goal,
                        backstory=bot.backstory,
                        rationale=bot.rationale,
                        pre_instructions=bot.pre_instructions,
                        company_information=bot.company_information,
                        vector_info=bot.database,
                        permissions=bot.permissions,
                    )
                    try:
                        await chatbot.configure(
                            app=app
                        )
                    except Exception as e:
                        self.logger.error(
                            f"Failed to configure chatbot '{chatbot.name}': {e}"
                        )
                elif bot.bot_type == 'agent':
                    self.logger.notice(
                        f"Unsupported kind of Agent '{bot.name}'..."
                    )
                    chatbot = None
                if chatbot:
                    self.add_bot(chatbot)
        self.logger.info(
            ":: Chatbots loaded successfully."
        )

    def create_bot(self, class_name: Any = None, name: str = None, **kwargs) -> AbstractBot:
        """Create a Bot and add it to the manager."""
        if class_name is None:
            class_name = Chatbot
        chatbot = class_name(**kwargs)
        chatbot.name = name
        self.add_bot(chatbot)
        if 'llm' in kwargs:
            llm = kwargs['llm']
            llm_name = llm.pop('name')
            model = llm.pop('model')
            llm = chatbot.load_llm(
                llm_name, model=model, **llm
            )
            chatbot.llm = llm
        return chatbot

    def add_bot(self, bot: AbstractBot) -> None:
        """Add a Bot to the manager."""
        self._bots[bot.name] = bot

    def get_bot(self, name: str) -> AbstractBot:
        """Get a Bot by name."""
        return self._bots.get(name)

    def remove_bot(self, name: str) -> None:
        """Remove a Bot by name."""
        del self._bots[name]

    def get_bots(self) -> Dict[str, AbstractBot]:
        """Get all Bots declared on Manager."""
        return self._bots

    def get_app(self) -> web.Application:
        """Get the app."""
        if self.app is None:
            raise RuntimeError("App is not set.")
        return self.app

    def setup(self, app: web.Application) -> web.Application:
        if isinstance(app, web.Application):
            self.app = app  # register the app into the Extension
        else:
            self.app = app.get_app()  # Nav Application
        # register signals for startup and shutdown
        self.app.on_startup.append(self.on_startup)
        self.app.on_shutdown.append(self.on_shutdown)
        # Add Manager to main Application:
        self.app['bot_manager'] = self
        ## Configure Routes
        router = self.app.router
        # Chat Information Router
        router.add_view(
            '/api/v1/chats',
            ChatHandler
        )
        router.add_view(
            '/api/v1/chat/{chatbot_name}',
            ChatHandler
        )
        # ChatBot Manager
        ChatbotHandler.configure(self.app, '/api/v1/bots')
        return self.app

    async def on_startup(self, app: web.Application) -> None:
        """On startup."""
        # configure all pre-configured chatbots:
        await self.load_bots(app)

    async def on_shutdown(self, app: web.Application) -> None:
        """On shutdown."""
        pass



from langchain_ollama import OllamaEmbeddings
from .base import BaseEmbed


class OllamaEmbed(BaseEmbed):
    """A wrapper class for Ollama embeddings."""
    model_name: str = "llama3"

    def _create_embedding(self, model_name: str = None, **kwargs):
        # Embedding Model:
        return OllamaEmbeddings(
            model_name=model_name or self.model_name,
            **kwargs
        )



from langchain_openai import (  # pylint: disable=E0401, E0611
    OpenAIEmbeddings
)
from .abstract import AbstractEmbed
from ...conf import OPENAI_API_KEY, OPENAI_ORGANIZATION


class OpenAIEmbed(AbstractEmbed):
    """A wrapper class for OpenAI embeddings."""
    model_name: str = "text-embedding-3-large"

    def _create_embedding(
        self,
        model_name: str = None,
        api_key: str = None,
        organization: str = None,
        **kwargs
    ):
        # Embedding
        return OpenAIEmbeddings(
            model=model_name or self.model_name,
            dimensions=kwargs.get('dimensions', 512),
            api_key=api_key or OPENAI_API_KEY,
            organization=organization or OPENAI_ORGANIZATION,
            max_retries=kwargs.get('max_retries', 4),
        )



from langchain_google_genai import (
    GoogleGenerativeAIEmbeddings
)
from .abstract import AbstractEmbed
from ...conf import GOOGLE_API_KEY

class GoogleEmbed(AbstractEmbed):
    """A wrapper class for VertexAI embeddings."""
    model_name: str = "models/embedding-001"

    def _create_embedding(self, model_name: str = None, api_key: str = None, **kwargs):
        # Embedding Model:
        api_key = api_key or GOOGLE_API_KEY
        return GoogleGenerativeAIEmbeddings(
            model=model_name or self.model_name,
            google_api_key=api_key,
            **kwargs
        )



from navigator.handlers.types import AppHandler
# Tasker:
from navigator.background import BackgroundQueue
from navigator_auth import AuthHandler
from parrot.manager import BotManager
from parrot.conf import STATIC_DIR
from parrot.handlers.bots import (
    FeedbackTypeHandler,
    ChatbotFeedbackHandler,
    PromptLibraryManagement,
    ChatbotUsageHandler,
    ChatbotSharingQuestion
)

class Main(AppHandler):
    """
    Main App Handler for Parrot Application.
    """
    app_name: str = 'Parrot'
    enable_static: bool = True
    enable_pgpool: bool = True
    staticdir: str = STATIC_DIR

    def configure(self):
        super(Main, self).configure()
        ### Auth System
        # create a new instance of Auth System
        auth = AuthHandler()
        auth.setup(self.app)
        # Tasker: Background Task Manager:
        tasker = BackgroundQueue(
            app=self.app,
            max_workers=5,
            queue_size=5
        )
        # Chatbot System
        self.bot_manager = BotManager()
        self.bot_manager.setup(self.app)

        # API of feedback types:
        self.app.router.add_view(
            '/api/v1/feedback_types/{feedback_type}',
            FeedbackTypeHandler
        )
        ChatbotFeedbackHandler.configure(self.app, '/api/v1/bot_feedback')
        # Prompt Library:
        PromptLibraryManagement.configure(self.app, '/api/v1/chatbots/prompt_library')
        # Questions (Usage handler, for sharing)
        ChatbotUsageHandler.configure(self.app, '/api/v1/chatbots/usage')
        self.app.router.add_view(
            '/api/v1/chatbots/questions/{sid}',
            ChatbotSharingQuestion
        )


    async def on_prepare(self, request, response):
        """
        on_prepare.
        description: Signal for customize the response while is prepared.
        """

    async def pre_cleanup(self, app):
        """
        pre_cleanup.
        description: Signal for running tasks before on_cleanup/shutdown App.
        """

    async def on_cleanup(self, app):
        """
        on_cleanup.
        description: Signal for customize the response when server is closing
        """

    async def on_startup(self, app):
        """
        on_startup.
        description: Signal for customize the response when server is started
        """
        app['websockets'] = []

    async def on_shutdown(self, app):
        """
        on_shutdown.
        description: Signal for customize the response when server is shutting down
        """
        pass



"""
Foundational base of every Chatbot and Agent in ai-parrot.
"""
from typing import Any, Union
from pathlib import Path, PurePath
import uuid
from aiohttp import web
# Navconfig
from navconfig import BASE_DIR
from navconfig.exceptions import ConfigError  # pylint: disable=E0611
from asyncdb.exceptions import NoDataFound
from ..utils import parse_toml_config
from ..conf import (
    default_dsn,
    EMBEDDING_DEFAULT_MODEL,
)
from ..models import ChatbotModel
from .abstract import AbstractBot

class Chatbot(AbstractBot):
    """Represents an Bot (Chatbot, Agent) in Navigator.

        Each Chatbot has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    def __init__(
        self,
        name: str = 'Nav',
        system_prompt: str = None,
        human_prompt: str = None,
        **kwargs
    ):
        """Initialize the Chatbot with the given configuration."""
        # Configuration File:
        self.config_file: PurePath = kwargs.get('config_file', None)
        # Other Configuration
        self.confidence_threshold: float = kwargs.get('threshold', 0.5)
        # Text Documents
        self.documents_dir = kwargs.get(
            'documents_dir',
            None
        )
        super().__init__(
            name=name,
            system_prompt=system_prompt,
            human_prompt=human_prompt,
            **kwargs
        )
        if isinstance(self.documents_dir, str):
            self.documents_dir = Path(self.documents_dir)
        if not self.documents_dir:
            self.documents_dir = BASE_DIR.joinpath('documents')
        if not self.documents_dir.exists():
            self.documents_dir.mkdir(
                parents=True,
                exist_ok=True
            )
        # define de Config File:
        if self.config_file:
            if isinstance(self.config_file, str):
                self.config_file = Path(self.config_file)
            if not self.config_file.exists():
                raise ConfigError(
                    f"Configuration file {self.config_file} not found."
                )

    def __repr__(self):
        return f"<ChatBot.{self.__class__.__name__}:{self.name}>"

    async def configure(self, app = None) -> None:
        """Load configuration for this Chatbot."""
        await super().configure(app)
        # Check if a Config File exists for this Bot instance:
        config_file = BASE_DIR.joinpath(
            'etc',
            'config',
            'chatbots',
            self.name.lower(),
            "config.toml"
        )
        if not config_file.exists():
            config_file = self.config_file or config_file
        if config_file.exists():
            self.logger.notice(
                f"Loading Bot {self.name} from config: {config_file.name}"
            )
            # Configure from the TOML file
            await self.from_config_file(config_file)
        elif (bot := await self.bot_exists(name=self.name, uuid=self.chatbot_id)):
            self.logger.notice(
                f"Loading Bot {self.name} from Database: {bot.chatbot_id}"
            )
            # Bot exists on Database, Configure from the Database
            await self.from_database(bot, config_file)
        else:
            raise ValueError(
                f'Bad configuration procedure for bot {self.name}'
            )

    def _from_bot(self, bot, key, config, default) -> Any:
        value = getattr(bot, key, None)
        file_value = config.get(key, default)
        return value if value else file_value

    def _from_db(self, botobj, key, default = None) -> Any:
        value = getattr(botobj, key, default)
        return value if value else default

    async def bot_exists(
        self,
        name: str = None,
        uuid: uuid.UUID = None
    ) -> Union[ChatbotModel, bool]:
        """Check if the Chatbot exists in the Database."""
        db = self.get_database('pg', dsn=default_dsn)
        async with await db.connection() as conn:  # pylint: disable=E1101
            ChatbotModel.Meta.connection = conn
            try:
                if self.chatbot_id:
                    try:
                        bot = await ChatbotModel.get(chatbot_id=uuid)
                    except Exception:
                        bot = await ChatbotModel.get(name=name)
                else:
                    bot = await ChatbotModel.get(name=self.name)
                if bot:
                    return bot
                else:
                    return False
            except NoDataFound:
                return False

    async def from_database(
        self,
        bot: Union[ChatbotModel, None] = None,
        config_file: PurePath = None
    ) -> None:
        """Load the Chatbot Configuration from the Database."""
        if not bot:
            db = self.get_database('pg', dsn=default_dsn)
            async with await db.connection() as conn:  # pylint: disable=E1101
                # import model
                ChatbotModel.Meta.connection = conn
                try:
                    if self.chatbot_id:
                        try:
                            bot = await ChatbotModel.get(chatbot_id=self.chatbot_id)
                        except Exception:
                            bot = await ChatbotModel.get(name=self.name)
                    else:
                        bot = await ChatbotModel.get(name=self.name)
                except NoDataFound:
                    # Fallback to File configuration:
                    raise ConfigError(
                        f"Chatbot {self.name} not found in the database."
                    )
        # Start Bot configuration from Database:
        if config_file and config_file.exists():
            file_config = await parse_toml_config(config_file)
            # Knowledge Base come from file:
            # Contextual knowledge-base
            self.kb = file_config.get('knowledge-base', [])
            if self.kb:
                self.knowledge_base = self.create_kb(
                    self.kb.get('data', [])
                )
        self.name = self._from_db(bot, 'name', default=self.name)
        self.chatbot_id = str(self._from_db(bot, 'chatbot_id', default=self.chatbot_id))
        self.description = self._from_db(bot, 'description', default=self.description)
        self.role = self._from_db(bot, 'role', default=self.role)
        self.goal = self._from_db(bot, 'goal', default=self.goal)
        self.rationale = self._from_db(bot, 'rationale', default=self.rationale)
        self.backstory = self._from_db(bot, 'backstory', default=self.backstory)
        # LLM Configuration:
        llm = self._from_db(bot, 'llm', default='vertexai')
        llm_config = self._from_db(bot, 'llm_config', default={})
        # Configuration of LLM:
        self.configure_llm(llm, llm_config)
        # Embedding Model Configuration:
        self.embedding_model : dict = self._from_db(
            bot, 'embedding_model', None
        )
        # Database Configuration:
        self._use_vector = bot.vector_store
        self._vector_store = bot.database
        self.configure_store()
        # after configuration, setup the chatbot
        if bot.system_prompt_template:
            self.system_prompt_template = bot.system_prompt_template
        self._define_prompt(
            config={}
        )
        # Last: permissions:
        _default = self.default_permissions()
        _permissions = bot.permissions
        self._permissions = {**_default, **_permissions}

    async def from_config_file(self, config_file: PurePath) -> None:
        """Load the Chatbot Configuration from the TOML file."""
        self.logger.debug(
            f"Using Config File: {config_file}"
        )
        file_config = await parse_toml_config(config_file)
        # getting the configuration from config
        self.config_file = config_file
        # basic config
        basic = file_config.get('chatbot', {})
        # Chatbot Name:
        self.name = basic.get('name', self.name)
        self.description = basic.get('description', self.description)
        self.role = basic.get('role', self.role)
        self.goal = basic.get('goal', self.goal)
        self.rationale = basic.get('rationale', self.rationale)
        self.backstory = basic.get('backstory', self.backstory)
        # Model Information:
        llminfo = file_config.get('llm')
        llm = llminfo.get('llm', 'VertexLLM')
        cfg = llminfo.get('config', {})
        # Configuration of LLM:
        self.configure_llm(llm, cfg)
        # Other models and embedding models:
        models = file_config.get('models', {})
        self.embedding_model = models.get(
            'embedding_model',
            {
                'model': EMBEDDING_DEFAULT_MODEL,
                'model_type': 'transformers'
            }
        )
        # pre-instructions
        instructions = file_config.get('pre-instructions')
        if instructions:
            self.pre_instructions = instructions.get('instructions', [])
        # Contextual knowledge-base
        self.kb = file_config.get('knowledge-base', [])
        if self.kb:
            self.knowledge_base = self.create_kb(
                self.kb.get('data', [])
            )
        database = file_config.get('database', {})
        vector_store = database.get('vector_store', False)
        if database or vector_store is True:
            self._use_database = True
        vector_db = database.pop('vector_database', None)
        # configure vector database:
        if vector_db:
            # Initialize the store:
            self.stores = []
            self.store = None
            self._use_vector = vector_store
            self._vector_store = database
            self.configure_store()
        # after configuration, setup the chatbot
        if 'template_prompt' in basic:
            self.template_prompt = basic.get('template_prompt')
        self._define_prompt(
            config=basic
        )
        # Last: permissions:
        permissions = file_config.get('permissions', {})
        _default = self.default_permissions()
        self._permissions = {**_default, **permissions}



"""
Executable Python REPL Tool.
"""
import io
import base64
import json
import matplotlib.pyplot as plt
from langchain_experimental.tools.python.tool import PythonAstREPLTool


class ExecutablePythonREPLTool(PythonAstREPLTool):
    """
    Executable Python REPL Tool.
    """
    def execute_code(self, code: str) -> str:
        """
        Execute the provided Python code and return the output.

        Args:
            code (str): The Python code to execute.

        Returns:
            str: The output of the executed code.
        """
        try:
            # Set up a namespace for execution
            namespace = {}
            exec(code, namespace)

            # Check if a plot was created
            if 'plt' in namespace:
                buf = io.BytesIO()
                plt.savefig(buf, format='png')
                plt.close()
                buf.seek(0)
                # Encode the image in base64
                # Encode the image in base64
                img_str = base64.b64encode(buf.read()).decode('utf-8')

                # Prepare the JSON output
                result = {
                    "image": {
                        "format": "png",
                        "base64": img_str
                    }
                }
                # Return both the code and the JSON result
                return f"**Code Executed**:\n```python\n{code}\n```\n\n**Result**:\n{json.dumps(result)}"
            else:
                return f"**Code Executed**:\n```python\n{code}\n```\n\n"

        except Exception as e:
            return f"Error executing code: {e}"

    def __call__(self, code: str) -> str:
        return self.execute_code(code)



AGENT_PROMPT = """
Your name is {name}.

{system_prompt_base}

**Answer the following questions as best you can. You have access to the following tools:**

- {tools}\n

Use these tools effectively to provide accurate and comprehensive responses:
{list_of_tools}

**Instructions:**
1. Understand the Query: Comprehend the user's request, especially if it pertains to events that may have already happened.
2. **Event Timing Validation**: For questions about recent events or events that may have happened already (like sporting events, conferences, etc.), if you're not confident that the event has happened, you must **use one of the web search tools** to confirm before making any conclusions.
3. Determine Confidence: If confident (90%+), provide the answer directly within the Thought process. If not confident, **always use a web search tool**.
4. Choose Tool: If needed, select the most suitable tool, using one of [{tool_names}].
5. Collect Information: Use the tool to gather data.
6. Analyze Information: Identify patterns, relationships, and insights.
7. Synthesize Response: Combine the information into a clear response.
8. Cite Sources: Mention the sources of the information.

** Your Style: **
- Maintain a professional and friendly tone.
- Be clear and concise in your explanations.
- Use simple language for complex topics to ensure user understanding.

To respond directly, use the following format:
```
Question: the input question you must answer.
Thought: Explain your reasoning.
Final Thought: Summarize your findings.
Final Answer: Provide a clear and structured answer to the original question with relevant details, always include the final answer of the tool in your final answer, also include your internal thoughts.
```


To use a tool, please use the following format:
```
Question: the input question you must answer.
Thought: Explain your reasoning, including whether you need to use a tool.
Action: the action to take, should be one of [{tool_names}].
- If using a tool: Specify the tool name (e.g., "Google Web Search") and the input.
Action Input: the input to the action.
Observation: the result of the action.
... (this Thought/Action/Action Input/Observation can repeat N times)
Final Thought: Summarize your findings.
Final Answer: Provide a clear and structured answer to the original question with relevant details, always include the final answer of the tool in your final answer, also include your internal thoughts.
Detailed Result: Include the detailed result from the tool here if applicable.
```

**Important**: Today is {today_date}, For any recent events you must **use a web search tool** to verify the outcome or provide accurate up-to-date information before concluding. Always prioritize using tools if you're unsure or if the event is recent.
- You must never contradict the given date.

Begin!

Question: {input}
{agent_scratchpad}
"""

SQL_AGENT_PROMPT = """
Your name is {name}. You are an agent designed to interact with a SQL database.
Given an input question, create a syntactically correct {dialect} query to run, then look at the results of the query and return the answer.

Use the following format:

Question: "Question here"
SQLQuery: "SQL Query to run"
SQLResult: "Result of the SQLQuery"
Answer: "Final answer here"


Unless the user specifies a specific number of examples they wish to obtain, always limit your query to at most {top_k} results.
You can order the results by a relevant column to return the most interesting examples in the database.
Never query for all the columns from a specific table, only ask for the relevant columns given the question.
You have access to tools for interacting with the database.

**Also you has access to the following extra tools:**

{list_of_tools}

Only use the below tools. Only use the information returned by the below tools to construct your final answer.
You MUST double check your query before executing it. If you get an error while executing a query, rewrite the query and try again.

DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

To start you should ALWAYS look at the tables in the database to see what you can query.
Do NOT skip this step.
Then you should query the schema of the most relevant tables.
Begin!

Question: {input}
{agent_scratchpad}
"""



from navconfig import config
from navconfig.logging import logging
from langchain_anthropic import ChatAnthropic  # pylint: disable=E0401, E0611
from langchain_core.prompts import ChatPromptTemplate
from .abstract import AbstractLLM

logging.getLogger(name='anthropic').setLevel(logging.WARNING)

class Anthropic(AbstractLLM):
    """Anthropic.

    Interact with Anthropic Language Model.

    Returns:
        _type_: an instance of Anthropic (Claude) LLM Model.
    """
    model: str = 'claude-3-opus-20240229'
    embed_model: str = None
    max_tokens: int = 1024
    supported_models: list = [
        'claude-3-opus-20240229',
        'claude-3-sonnet-20240229',
        'claude-3-haiku-20240307',
        'claude-3-5-sonnet-20240620',
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = kwargs.get("model", 'claude-3-opus-20240229')
        self._api_key = kwargs.pop('api_key', config.get('ANTHROPIC_API_KEY'))
        args = {
            "temperature": self.temperature,
            "max_retries": 4,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "verbose": True,
        }
        self._llm = ChatAnthropic(
            model_name=self.model,
            api_key=self._api_key,
            **args
        )



from langchain_community.utilities.bing_search import BingSearchAPIWrapper
from langchain.tools import BaseTool


class BingSearchTool(BaseTool):
    """Microsoft Bing Search Tool."""
    name: str = "Bing Search"
    description: str = "Search the web using Microsoft Bing Search, conduct searches using Bing for alternative perspectives and sources."

    def _run(self, query: str) -> dict:
        """Run the Bing Search Tool."""
        bing = BingSearchAPIWrapper(k=5)
        return bing.results(query=query, num_results=5)



from .abstract import AbstractStore
from .empty import EmptyStore

supported_stores = {
    'chroma': 'ChromaStore',
    'duck': 'DuckDBStore',
    'milvus': 'MilvusStore',
    'qdrant': 'QdrantStore',
    'postgres': 'PgvectorStore',
    'faiss': 'FaissStore',
}



# cython: language_level=3, embedsignature=True, boundscheck=False, wraparound=True, initializedcheck=False
# Copyright (C) 2018-present Jesus Lara
#
cdef class ParrotError(Exception):
    """Base class for other exceptions"""
    pass

cdef class ConfigError(ParrotError):
    pass



from langchain_huggingface import HuggingFaceEmbeddings
from .base import BaseEmbed


class TransformersEmbed(BaseEmbed):
    """A wrapper class for Transformers embeddings."""
    model_name: str = "all-MiniLM-L6-v2"
    #  "sentence-transformers/all-mpnet-base-v2"

    def _create_embedding(self, model_name: str = None, **kwargs):
        # Embedding Model:
        device = self._get_device()
        model_args = {
            **self.model_kwargs,
            'device': device,
        }
        return HuggingFaceEmbeddings(
            model_name=model_name or self.model_name,
            model_kwargs=model_args,
            encode_kwargs=self.encode_kwargs
        )



from langchain.tools import Tool
# Tools:
from langchain.chains.llm_math.base import LLMMathChain
from langchain_community.utilities import SerpAPIWrapper
from langchain_groq import ChatGroq
from ..conf import (
    SERPAPI_API_KEY,
    GROQ_API_KEY
)

class SearchTool:
    """Search Tool."""
    name: str = "Web Search"
    def __new__(cls, *args, **kwargs):
        search = SerpAPIWrapper(
            serpapi_api_key=kwargs.get('serpapi_api_key', SERPAPI_API_KEY)
        )
        tool = Tool(
            name="Web Search",
            func=search.run,
            description="""
            useful for when you need to answer questions about current events or general knowledge. Input should be a search query.
            """,
            coroutine=search.arun
        )
        tool.name = "Web Search"
        return tool


class MathTool:
    """Math Tool."""
    name: str = "Math Calculator"
    def __new__(cls, *args, **kwargs):
        groq = ChatGroq(
            model="llama-3.3-70b-versatile",
            temperature=0.1,
            api_key=kwargs.get('GROQ_API_KEY', GROQ_API_KEY)
        )
        llm = kwargs.get('llm', groq)
        math_chain = LLMMathChain.from_llm(
            llm=llm,
            verbose=True
        )
        tool = Tool(
            name="Math Calculator",
            func=math_chain.run,
            description="""
            useful for when you need to solve math problems or perform mathematical calculations. Input should be a math equation or a mathematical expression.
            """,
            coroutine=math_chain.arun
        )
        tool.name = "Math Calculator"
        return tool



from .chatbot import Chatbot

class Cody(Chatbot):
    """Represents a Python expert in Navigator.

        Each expert has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    name: str = 'Cody'
    company_information: dict = {
        'company': 'T-ROC Global',
        'company_website': 'https://www.trocglobal.com',
        'contact_email': 'communications@trocglobal.com',
        'contact_form': 'https://www.surveymonkey.com/r/TROC_Suggestion_Box'
    }
    role: str = 'Python Expert'
    goal = 'Provide useful information about Python programming to employees.'



from typing import Optional, Any
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_community.tools import WikipediaQueryRun
from langchain_community.utilities import WikipediaAPIWrapper
from langchain_community.tools.wikidata.tool import WikidataAPIWrapper, WikidataQueryRun
from langchain.tools import BaseTool


class WikipediaTool(BaseTool):
    """Tool that searches the Wikipedia API."""
    name = "Wikipedia"
    description: str = (
        "Access detailed and verified information from Wikipedia. "
        "Useful for searching Wikipedia for general information. "
        "Useful for when you need to answer general questions about "
        "people, places, companies, facts, historical events, or other subjects. "
        "Input should be a search query."
    )
    search: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.search = WikipediaQueryRun(
            api_wrapper=WikipediaAPIWrapper()
        )

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the Wikipedia tool."""
        return self.search.run(query)

class WikidataTool(BaseTool):
    """Tool that searches the Wikidata API."""
    name: str = "Wikidata"
    description: str = (
        "Fetch structured data from WikiData for precise and factual details. "
        "Useful for when you need to answer general questions about "
        "people, places, companies, facts, historical events, or other subjects. "
        "Input should be the exact name of the item you want information about or a Wikidata QID."
    )
    search: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.search = WikidataQueryRun(
            api_wrapper=WikidataAPIWrapper()
        )

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the Wikidata tool."""
        return self.search.run(query)



from crewai_tools import WebsiteSearchTool
from crewai_tools import BaseTool


class WebSearchTool(BaseTool):
    """Search on any Website using WebSiteSearch."""
    name: str = "WebSite Search"
    description: str = "Search on a website using WebsiteSearchTool"

    def _run(self, query: str, **kwargs) -> dict:
        """Run the Google Trends Tool."""
        search = WebsiteSearchTool(
            config={
                "llm": {
                    "provider": "google",
                    "config": {
                        "model": "models/gemini-pro",
                        "temperature": 0.4,
                        "top_p": 1,
                        "stream": True
                    }
                },
                "embedder": {
                    "provider": "google",
                    "config": {
                        "model": "models/embedding-001",
                        "task_type": "retrieval_document",
                    }
                }
            }
        )
        return search.run(query, **kwargs)



from .chatbot import Chatbot


class OddieBot(Chatbot):
    """Represents an agent in Navigator.

        Each agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    name: str = 'Oddie'
    company: str = 'T-ROC Global'
    company_website: str = 'https://www.trocglobal.com'
    contact_information = 'communications@trocglobal.com'
    contact_form = 'https://www.surveymonkey.com/r/TROC_Suggestion_Box'
    role = "Odoo and ERP Specialist and Odoo Programmer"
    goal = "To provide information and support on Odoo and ERP systems, help with troubleshooting, and answer any questions you may have about any Odoo and ERP systems implementation."
    specialty_area = 'Bring useful information about Odoo ERP, documentation, usage, samples, etc.'



from asyncdb import AsyncDB
from navigator.views import (
    ModelView,
    BaseView,
    FormModel
)
from parrot.conf import (
    BIGQUERY_CREDENTIALS,
    BIGQUERY_PROJECT_ID,
)
from ..models import (
    ChatbotModel,
    ChatbotUsage,
    PromptLibrary,
    ChatbotFeedback,
    FeedbackType
)


class PromptLibraryManagement(ModelView):
    """
    PromptLibraryManagement.
    description: PromptLibraryManagement for Parrot Application.
    """

    model = PromptLibrary
    name: str = "Prompt Library Management"
    path: str = '/api/v1/prompt_library'
    pk: str = 'prompt_id'

    async def _set_created_by(self, value, column, data):
        if not value:
            return await self.get_userid(session=self._session)
        return value


class ChatbotUsageHandler(ModelView):
    """
    ChatbotUsageHandler.
    description: ChatbotUsageHandler for Parrot Application.
    """

    model = ChatbotUsage
    driver: str = 'bigquery'
    credentials: dict = {
        "credentials": BIGQUERY_CREDENTIALS,
        "project_id": BIGQUERY_PROJECT_ID,
    }
    name: str = "Chatbot Usage"
    path: str = '/api/v1/chatbots/usage'
    pk: str = 'sid'


class ChatbotSharingQuestion(BaseView):
    """
    ChatbotSharingQuestion.
    description: ChatbotSharingQuestion for Parrot Application.
    """

    def get_connection(self):
        params = {
            "credentials": BIGQUERY_CREDENTIALS,
            "project_id": BIGQUERY_PROJECT_ID,
        }
        return AsyncDB(
            'bigquery',
            params=params
        )

    async def get(self):
        qs = self.get_arguments(self.request)
        sid = qs.get('sid', None)
        if not sid:
            return self.error(
                response={
                    "message": "You need to Provided a ID of Question"
                },
                status=400
            )
        db = self.get_connection()
        try:
            async with await db.connection() as conn:  #pylint: disable=E1101
                ChatbotUsage.Meta.connection = conn
                # Getting a SID from sid
                question = await ChatbotUsage.get(sid=sid)
                if not question:
                    return self.error(
                        response={
                            "message": "Question not found"
                        },
                        status=404
                    )
                return self.json_response(
                    {
                        "chatbot": question.chatbot_id,
                        "question": question.question,
                        "answer": question.response,
                        "at": question.used_at
                    }
                )
        except Exception as e:
            return self.error(
                response={
                    "message": f"Error on Chatbot Sharing Question: {e}"
                },
                status=400
            )



class FeedbackTypeHandler(BaseView):
    """
    FeedbackTypeHandler.
    description: FeedbackTypeHandler for Parrot Application.
    """

    async def get(self):
        qs = self.get_arguments(self.request)
        category = qs.get('feedback_type', 'good').capitalize()
        feedback_list = FeedbackType.list_feedback(category)
        return self.json_response({
            "feedback": feedback_list
        })

# Manage Feedback:
class ChatbotFeedbackHandler(FormModel):
    """
    ChatbotFeedbackHandler.
    description: ChatbotFeedbackHandler for Parrot Application.
    """
    model = ChatbotFeedback
    path: str = '/api/v1/bot_feedback'

    def get_connection(self):
        params = {
            "credentials": BIGQUERY_CREDENTIALS,
            "project_id": BIGQUERY_PROJECT_ID,
        }
        return AsyncDB(
            'bigquery',
            params=params,
            force_closing=False
        )

    async def post(self):
        feedback = await self.validate_payload()
        if not feedback:
            return self.error(
                response={
                    "message": "Error on Bot Feedback"
                },
                status=400
            )
        db = self.get_connection()
        try:
            async with await db.connection() as conn:  #pylint: disable=E1101
                data = feedback.to_dict()
                # convert to string (bigquery uses json.dumps to convert to string)
                data['sid'] = str(data['sid'])
                data['chatbot_id'] = str(data['chatbot_id'])
                data['expiration_timestamp'] = str(data['expiration_timestamp'])
                data['feedback_type'] = feedback.feedback_type.value
                # writing directly to bigquery
                await conn.write(
                    [data],
                    table_id=ChatbotFeedback.Meta.name,
                    dataset_id=ChatbotFeedback.Meta.schema,
                    use_streams=False,
                    use_pandas=False
                )
                return self.json_response({
                    "message": "Bot Feedback Submitted, Thank you for your feedback!.",
                    "question": f"Question of ID: {feedback.sid} for bot {feedback.chatbot_id}"
                }, status=201)
        except Exception as e:
            return self.error(
                response={
                    "message": f"Error on Bot Feedback: {e}"
                },
                status=400
            )


class ChatbotHandler(ModelView):
    """
    ChatbotHandler.
    description: ChatbotHandler for Parrot Application.
    """

    model = ChatbotModel
    name: str = "Chatbot Management"
    pk: str = 'chatbot_id'


    async def _set_created_by(self, value, column, data):
        return await self.get_userid(session=self._session)



from typing import List, Dict, Any, Optional
from collections.abc import Callable
from langchain_core.retrievers import BaseRetriever
from langchain_core.callbacks import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from pydantic import Field

class MultiVectorStoreRetriever(BaseRetriever):
    """
    This aggregator retriever queries multiple vector stores
    and merges the results into a single list.
    """

    # Define class attributes with default values
    stores: List[Any] = Field(description="List of vector stores that provide an as_retriever() method")
    search_kwargs: Dict[str, Any] = Field(default_factory=dict, description="Dict to pass to each store's retrieve method (like 'k' for top_k)")
    chain_type: str = Field(default="stuff", description="Chain type for the retriever")
    search_type: str = Field(default="similarity", description="Search type (similarity, mmr)")
    metric_type: str = Field(default="COSINE", description="Similarity metric (COSINE, EUCLIDEAN, DOT_PRODUCT)")

    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True

    def __init__(
        self,
        stores: List[Any],
        metric_type: str = 'COSINE',
        chain_type: str = 'stuff',
        search_type: str = 'similarity',
        search_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        """
        Initialize the MultiVectorStoreRetriever.

        Args:
            stores: List of vector stores that provide an as_retriever() method
            metric_type: Similarity metric (COSINE, EUCLIDEAN, DOT_PRODUCT)
            chain_type: Chain type for the retriever
            search_type: Search type (similarity, mmr)
            search_kwargs: Dict to pass to each store's retrieve method (like 'k' for top_k)
        """
        # Initialize with default values
        search_kwargs = search_kwargs or {}

        # Call super().__init__ with all properties
        super().__init__(
            stores=stores,
            search_kwargs=search_kwargs,
            chain_type=chain_type,
            search_type=search_type,
            metric_type=metric_type,
            **kwargs
        )

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Get documents relevant to the query.
        Implements the required method from BaseRetriever.

        Args:
            query: Query string
            run_manager: CallbackManager for the run

        Returns:
            List of relevant documents
        """
        all_results = []
        for store in self.stores:
            try:
                retriever = store.as_retriever(
                    search_type=self.search_type,
                    search_kwargs=self.search_kwargs,
                )
                # Pass the run_manager to the sub-retrievers if they support it
                if hasattr(retriever, "_get_relevant_documents"):
                    callback_manager = run_manager.get_child()
                    docs = retriever._get_relevant_documents(query, run_manager=callback_manager)
                else:
                    docs = retriever.get_relevant_documents(query)
                all_results.extend(docs)
            except Exception as e:
                # Log the error but continue with other stores
                run_manager.on_retriever_error(f"Error retrieving from store: {str(e)}")
                continue
        return all_results

    async def _aget_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        """
        Asynchronously get documents relevant to the query.
        Implements the required async method from BaseRetriever.

        Args:
            query: Query string
            run_manager: CallbackManager for the run

        Returns:
            List of relevant documents
        """
        all_results = []
        for store in self.stores:
            try:
                retriever = store.as_retriever(
                    search_type=self.search_type,
                    search_kwargs=self.search_kwargs,
                )
                # Pass the run_manager to the sub-retrievers if they support it
                if hasattr(retriever, "_aget_relevant_documents"):
                    callback_manager = run_manager.get_child()
                    docs = await retriever._aget_relevant_documents(query, run_manager=callback_manager)
                else:
                    docs = await retriever.aget_relevant_documents(query)
                all_results.extend(docs)
            except Exception as e:
                # Log the error but continue with other stores
                run_manager.on_retriever_error(f"Error retrieving from store: {str(e)}")
                continue
        return all_results



# cython: language_level=3, embedsignature=True, boundscheck=False, wraparound=True, initializedcheck=False
# Copyright (C) 2018-present Jesus Lara
#
import time
import orjson
from typing import (
    Dict
)
from collections.abc import Sequence
from decimal import Decimal
from numpy import int64, ndarray
from datetime import timezone
from cpython cimport datetime
from zoneinfo import ZoneInfo
from uuid import UUID
from dateutil.parser import parse, ParserError



cpdef object strtobool(str val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0', 'null'):
        return False
    else:
        raise ValueError(
            f"invalid truth value for {val}"
        )


cpdef datetime.date to_date(object value, str mask = "%Y-%m-%d %H:%M:%S", str tz = None):
    if isinstance(value, datetime.datetime):
        return value
    else:
        try:
            result = datetime.datetime.strptime(str(value), mask)
            if tz is not None:
                zone = ZoneInfo(key=tz)
                result = result.replace(tzinfo=zone)
            return result
        except (TypeError, ValueError, AttributeError):
            return parse(str(value))


cpdef datetime.time to_time(object value, str mask = "%H:%M:%S"):
    if value == 0:
        return datetime.datetime.now().replace(
            hour=0, minute=0, second=0, microsecond=0
        )
    if isinstance(value, datetime.time):
        return value
    elif isinstance(value, datetime.datetime):
        return value.time()
    else:
        if len(str(value)) < 6:
            value = str(value).zfill(6)
        try:
            return datetime.datetime.strptime(str(value), mask)
        except ValueError:
            return datetime.datetime.strptime(str(value), "%H:%M:%S")


cpdef datetime.datetime to_datetime(object value, str mask = "%Y-%m-%d %H:%M:%S"):
    if isinstance(value, datetime.datetime):
        return value
    elif isinstance(value, list):
        dt = to_date(value[0], mask=mask[0])
        mt = to_time(value[1], mask=mask[1]).time()
        return datetime.datetime.combine(dt, mt)
    else:
        if value is None:
            return datetime.datetime.now().replace(
                hour=0, minute=0, second=0, microsecond=0
            )
        else:
            return datetime.datetime.strptime(str(value), mask)


cpdef datetime.datetime epoch_to_date(object value, str tz = None):
    if value is None:
        return None
    if len(str(value)) == 10:
        s = value
    elif len(str(value)) == 19:
        s = value / 1e9
    else:
        s, _ = divmod(value, 1000.0)
    if tz is not None:
        zone = ZoneInfo(key=tz)
    else:
        zone = timezone.utc
    return datetime.datetime.fromtimestamp(s, zone)


cpdef object to_boolean(object value):
    if isinstance(value, bool):
        return value
    elif value is None:
        return False
    else:
        try:
            return strtobool(value)
        except ValueError:
            return False


cpdef object is_date(object value):
        if isinstance(value, (datetime.datetime, time, datetime.time, datetime.timedelta)):
            return True
        else:
            try:
                val = parse(value)
                if val:
                    return True
                return False
            except (ParserError, ValueError):
                return False

cdef class Entity:
    """Entity.
    Used to convert entities (string, number, dates) to appropiated string on SQL queries.
    """

    @classmethod
    def is_integer(cls, _type):
        return _type in (int, int64)

    @classmethod
    def is_number(cls, _type):
        return _type in (int, int64, float, Decimal, bytes, bool)

    @classmethod
    def is_string(cls, _type):
        return isinstance(_type, (
            str,
            datetime.datetime,
            datetime.time,
            datetime.timedelta,
            UUID
        ))

    @classmethod
    def is_date(cls, _type):
        return _type in (datetime.datetime, time, datetime.time, datetime.timedelta)

    @classmethod
    def is_array(cls, t):
        return isinstance(t,(list, dict, Sequence, ndarray))

    @classmethod
    def is_bool(cls, _type):
        return isinstance(_type, bool)

    @classmethod
    def toSQL(cls, value, _type, dbtype: str = None):
        v = f"{value!r}" if Entity.is_date(_type) else value
        v = f"{value!s}" if Entity.is_string(_type) and value is not None else v
        v = value if Entity.is_number(_type) else v
        v = str(value) if isinstance(value, UUID) else v
        # json conversion
        v = orjson.dumps(value) if _type in [dict, Dict] else v
        v = f"{value!s}" if dbtype == "array" and value is not None else v
        # formatting htstore column
        v = (
            ",".join({"{}=>{}".format(k, v) for k, v in value.items()})
            if isinstance(value, dict) and dbtype == "hstore"
            else v
        )
        v = "NULL" if (value in ["None", "null"]) else v
        v = "NULL" if value is None else v
        return v

    @classmethod
    def escapeLiteral(cls, value, _type, dbtype: str = None):
        v = value if value != "None" or value is not None else ""
        v = f"{value!r}" if Entity.is_string(_type) else v
        v = value if Entity.is_number(_type) else f"{value!r}"
        v = f"array{value!s}" if dbtype == "array" else v
        v = f"{value!r}" if dbtype == "hstore" else v
        v = value if (value in ["None", "null", "NULL"]) else v
        return v

    @classmethod
    def escapeString(cls, value):
        v = value if value != "None" else ""
        v = str(v).replace("'", "''")
        v = "'{}'".format(v) if Entity.is_string(type(value)) else v
        return v

    @classmethod
    def quoteString(cls, value):
        v = value if value != "None" else ""
        v = "'{}'".format(v) if type(v) == str else v
        return v

class SafeDict(dict):
    """
    SafeDict.

    Allow to using partial format strings

    """

    def __missing__(self, key):
        """Missing method for SafeDict."""
        return "{" + key + "}"



from collections.abc import Callable
from typing import Optional, Union
import duckdb
from langchain.docstore.document import Document
from langchain.memory import VectorStoreRetrieverMemory
from langchain_community.vectorstores import DuckDB
from .abstract import AbstractStore


class DuckDBStore(AbstractStore):
    """DuckDB Store Class.

    Using DuckDB as Document Vector Store.

    """
    default_config: dict ={
        "enable_external_access": "false",
        "autoinstall_known_extensions": "false",
        "autoload_known_extensions": "false"
    }

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )
        self.credentials = {
            "database": self.database,
        }

        config: dict = kwargs.pop("config", {})
        self.config = {
            **self.default_config,
            **config
        }

    async def connection(self, alias: str = None):
        """Connection to DuckDB.

        Args:
            alias (str): Database alias.

        Returns:
            Callable: DuckDB connection.

        """
        self._connection = duckdb.connect(**self.credentials, config=self.config)
        self._connected = True
        return self._connection

    async def disconnect(self) -> None:
        """
        Closing the Connection on DuckDB
        """
        try:
            if self._connection:
                self._connection.close()
        except Exception as err:
            raise RuntimeError(
                message=f"{__name__!s}: Closing Error: {err!s}"
            ) from err
        finally:
            self._connection = None
            self._connected = False

    def get_vector(
        self,
        collection: Union[str, None] = None,
        embedding: Optional[Callable] = None,
        metadata_field: str = 'id',
        text_field: str = 'text',
        vector_key: str = 'vector',
    ) -> DuckDB:

        if not collection:
            collection = self.collection_name
        if embedding is not None:
            _embed_ = embedding
        else:
            _embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        return DuckDB(
            connection=self._connection,
            table_name=collection,
            embedding=_embed_,
            vector_key=vector_key,
            text_key=text_field,
            id_key=metadata_field
        )


    async def add_texts(self, objects: list, collection: str = None):
        """
        Add Texts to DuckDB
        """
        async with self:
            store = self.get_vector(collection=collection)
            store.add_texts(objects)
        return True

    async def similarity_search(
        self,
        query: str,
        collection: Union[str, None] = None,
        embedding: Optional[Callable] = None,
        limit: int = 2,
    ) -> list:
        if collection is None:
            collection = self.collection_name
        async with self:
            vector_db = self.get_vector(collection=collection, embedding=embedding)
            return await vector_db.asimilarity_search(query, k=limit)

    def memory_retriever(
        self,
        documents: Optional[list] = None,
        num_results: int  = 5
    ) -> VectorStoreRetrieverMemory:
        if not documents:
            documents = []
        vectordb = DuckDB.from_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            connection=self._connection,
        )
        retriever = DuckDB.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)

    async def from_documents(self, documents: list[Document], collection: str = None, **kwargs):
        """
        Save Documents as Vectors in DuckDB.
        """
        if not collection:
            collection = self.collection_name
        vectordb = await DuckDB.afrom_documents(
            documents,
            embedding=self._embed_.embedding,
            connection=self._connection,
        )
        return vectordb

    async def add_documents(self, documents: list[Document], collection: str = None, **kwargs):
        """
        Add Documents as Vectors in DuckDB.
        """
        if not collection:
            collection = self.collection_name
        vectordb = self.get_vector(collection=collection)
        result = await vectordb.aadd_documents(
            documents=documents
        )
        return result



"""
Tools infrastructure for building Agents.
"""
# Abstract Agent interface
from .abstract import AbstractTool
from .basic import SearchTool, MathTool
from .duck import DuckDuckGoSearchTool, DuckDuckGoRelevantSearch

# from langchain_community.tools.yahoo_finance_news import YahooFinanceNewsTool
# from langchain_community.tools import YouTubeSearchTool
# from langchain_community.agent_toolkits import O365Toolkit
# from navconfig import config
# # from .wikipedia import WikipediaTool, WikidataTool
# from .asknews import AskNewsTool

# from .weather import OpenWeather, OpenWeatherMapTool
# from .google import GoogleLocationFinder, GoogleSiteSearchTool, GoogleSearchTool
# from .zipcode import ZipcodeAPIToolkit
# from .bing import BingSearchTool
# from .stack import StackExchangeTool



from collections.abc import Callable
from typing import Optional, Union
from uuid import uuid4
import logging
from langchain.docstore.document import Document
from langchain.memory import VectorStoreRetrieverMemory
import chromadb
from langchain_chroma import Chroma
from .abstract import AbstractStore
from ..conf import CHROMADB_HOST, CHROMADB_PORT


logging.getLogger('chromadb').setLevel(logging.INFO)


class ChromaStore(AbstractStore):
    """Chroma DB Store Class.

    Using Chroma as Document Vector Store.

    """

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )
        self.database_path: str = kwargs.pop('database_path', 'chroma.db')
        self._ephemeral: bool = kwargs.pop('ephemeral', False)
        self._local: bool = kwargs.pop('local', False)
        self.host = kwargs.pop("host", CHROMADB_HOST)
        self.port = kwargs.pop("port", CHROMADB_PORT)
        self._collection = None

    async def connection(self):
        """Connection to ChromaDB.

        Args:
            alias (str): Database alias.

        Returns:
            Callable: ChromaDB connection.

        """
        if self._ephemeral:
            self._connection = chromadb.Client()
        elif self._local:
            self._connection = chromadb.PersistentClient(
                path=self.database_path,
                database=self.database,
            )
        else:
            # Client-Server Connection:
            self._connection = chromadb.HttpClient(
                host=self.host,
                port=self.port,
                database=self.database,
            )
        self._collection = self._connection.get_or_create_collection(self.collection_name)
        self._connected = True
        return self._connection

    async def disconnect(self) -> None:
        """
        Closing the Connection on ChromaDB
        """
        self._connection = None
        self._connected = False

    def get_vector(
        self,
        collection: Union[str, None] = None,
        embedding: Optional[Callable] = None,
    ) -> Chroma:

        if not collection:
            collection = self.collection_name
        if embedding is not None:
            _embed_ = embedding
        else:
            _embed_ = self._embed_ or self.create_embedding(
                embedding_model=self.embedding_model
            )
        return Chroma(
            collection_name=self.collection_name,
            embedding_function=_embed_.embedding,
            client=self._connection,
            create_collection_if_not_exists=True,
        )

    async def from_documents(self, documents: list[Document], collection: str = None, **kwargs):
        """
        Save Documents as Vectors in Chroma.
        """
        vectordb = await Chroma.afrom_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            connection=self._connection,
        )
        return vectordb


    async def add_texts(self, objects: list, collection: str = None):
        """
        Add Texts to ChromaDB
        """
        async with self:
            collection = self._connection.get_or_create_collection(collection)
            for i, doc in enumerate(objects):
                collection.add(ids=[str(i)], documents=[doc])
        return True

    async def add_documents(
        self,
        documents: list,
        collection: str = None,
        embedding: Optional[Callable] = None,
    ) -> bool:
        """Add Documents to ChromaDB"""

        if collection is None:
            collection = self.collection_name

        async with self:
            collection_obj = self._connection.get_or_create_collection(collection)
            uuids = [str(uuid4()) for _ in range(len(documents))]
            vector_db = self.get_vector(collection=collection, embedding=embedding)
            await vector_db.aadd_documents(documents=documents, ids=uuids)

        return True


    async def update_documents(
        self,
        documents: list,
        collection: str = None,
        embedding: Optional[Callable] = None,
    )   -> bool:
        """
        Update Documents to ChromaDB
        """
        async with self:
            collection = self._connection.get_or_create_collection(collection)
            vector_db = self.get_vector(collection=collection, embedding=embedding)
            # Split the documents into ids and documents
            if all('id' in doc for doc in documents):
                ids = [doc.pop('id') for doc in documents]
                vector_db.update_documents(documents=documents, ids=ids)
                return True
            return False

    async def similarity_search(
        self,
        query: str,
        collection: Union[str, None] = None,
        embedding: Optional[Callable] = None,
        limit: int = 2,
        filter: Optional[dict] = None,
    ) -> list:
        if collection is None:
            collection = self.collection_name
        async with self:
            vector_db = self.get_vector(collection=collection, embedding=embedding)
            return vector_db.similarity_search(query, k=limit, filter=filter)

    def memory_retriever(
        self,
        documents: Optional[list] = None,
        num_results: int  = 5
    ) -> VectorStoreRetrieverMemory:
        if not documents:
            documents = []
        vectordb = Chroma.from_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            connection=self._connection,
        )
        retriever = Chroma.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)
